
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <direct.h>

#define VPK_SIGN1		0x04034B50
#define VPK_SIGN2		0x02014B50
#define VPK_SIGN3		0x06054B50

#define BUFFER_SIZE		(64 * 1024 * 1024)

#pragma pack(push, 1)
struct vpkhdr1_t
{
	unsigned char unk[14];
	unsigned filelen;
	unsigned filelen2;
	unsigned namelen;
};

struct vpkhdr2_t
{
	unsigned char unk[24];
	unsigned short namelen;
	unsigned len;
	unsigned char unk2[12];
};

struct vpkhdr3_t
{
	unsigned char unk[18];
};
#pragma pack(pop)

char vpkfile[_MAX_PATH];
char gamedir[_MAX_PATH];
int gamedirlen;
FILE *filehandle;

char buffer[BUFFER_SIZE];

void fatal(const char *reason)
{
	printf("\nfatal: %s\n", reason);
	if (filehandle) {
		int ofs = ftell(filehandle);
		fseek(filehandle, 0, SEEK_END);
		int size = ftell(filehandle);
		fclose(filehandle);
		int left = size - ofs;
		printf("\noffset %X\nprocessed %d out of %d bytes, %d left\n", ofs, ofs, size, left);
	}
	exit(1);
}

void sanitize(char *path)
{
	while (*path) {
		if (*path == '/') {
			*path = '\\';
		}
		path++;
	}
}

void makegamedir()
{
	strcpy(gamedir, vpkfile);
	if (char *slash = strrchr(gamedir, '\\')) {
		*(slash + 1) = '\0';
	} else {
		gamedir[0] = '\0';
	}
	gamedirlen = strlen(gamedir);
}

void writefile(const char *name, int len, FILE *src)
{
	int nl = strlen(name);
	if (gamedirlen + nl >= _MAX_PATH) {
		fatal("path too long");
	}
	char fullname[_MAX_PATH];
	strcpy(fullname, gamedir);
	strcat(fullname, name);
	if (name[nl - 1] == '/' || name[nl - 1] == '\\') {
		_mkdir(name);
	} else {
		FILE *dst = fopen(fullname, "wb");
		if (!dst) {
			fatal("can't create file");
		}
		while (len > 0) {
			int chunksize = len;
			if (chunksize > BUFFER_SIZE) {
				chunksize = BUFFER_SIZE;
			}
			fread(buffer, 1, chunksize, src);
			fwrite(buffer, 1, chunksize, dst);
			len -= chunksize;
		}
		fclose(dst);
	}
}

void extract()
{
	filehandle = fopen(vpkfile, "rb");
	if (!filehandle) {
		fatal("can't open file");
	}
	for (;;) {
		unsigned sign;
		int n = fread(&sign, 1, sizeof(sign), filehandle);
		if (!n) {
			break;
		}
		if (n < sizeof(sign)) {
			fatal("unexpected end of file");
		}
		if (sign == VPK_SIGN1) {
			vpkhdr1_t hdr;
			fread(&hdr, 1, sizeof(hdr), filehandle);
			if (hdr.filelen != hdr.filelen2) {
				fatal("size mismatch, maybe compressed file");
			}
			if (hdr.namelen >= _MAX_PATH) {
				fatal("name too long");
			}
			char name[_MAX_PATH];
			fread(name, 1, hdr.namelen, filehandle);
			name[hdr.namelen] = '\0';
			printf("%s\n", name);
			writefile(name, hdr.filelen, filehandle);
		} else if (sign == VPK_SIGN2) {
			vpkhdr2_t hdr;
			fread(&hdr, 1, sizeof(hdr), filehandle);
			fseek(filehandle, hdr.namelen, SEEK_CUR);
			fseek(filehandle, hdr.len, SEEK_CUR);
		} else if (sign == VPK_SIGN3) {
			vpkhdr3_t hdr;
			fread(&hdr, 1, sizeof(hdr), filehandle);
		} else {
			fatal("signature error");
		}
	}
	printf("\nall done!\n");
	fclose(filehandle);
	filehandle = nullptr;
}

int main(int argc, char **argv)
{
	if (argc < 2) {
		printf("usage:\n");
		printf("  vpkextract vpk-file-path\n");
		return 0;
	}
	strncpy(vpkfile, argv[1], _MAX_PATH);
	if (vpkfile[_MAX_PATH - 1]) {
		fatal("path too long");
	}
	sanitize(vpkfile);
	makegamedir();
	extract();
	return 0;
}

